import 'dart:convert';

import 'package:backend_dart/application/service/dto/project_dto.dart';
import 'package:backend_dart/application/service/mapper/project_dto_mapper.dart';
import 'package:backend_dart/common/request_helper.dart';
import 'package:backend_dart/common/response_helpers.dart';
import 'package:backend_dart/common/validation.dart';
import 'package:backend_dart/domain/repository/project_repository.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_router/shelf_router.dart';

part 'project_service.g.dart'; // generated with 'pub run build_runner build'

class ProjectService {
  final ProjectRepository projects;
  final ProjectDtoMapper mapper = ProjectDtoMapper();
  ProjectService(this.projects);

  @Route.get('/')
  Future<Response> listProjects(Request request) async {
    return projects
        .findAll()
        .flatMap(mapper.listTo)
        .map((projects) => projects.map((project) => project.toJson()).toList())
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok(jsonEncode(right)))
        .run();
  }

  @Route.get('/<projectId>')
  Future<Response> fetchProject(Request request, String projectId) async {
    return projects
        .findById(projectId)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.post('/')
  Future<Response> createProject(Request request) async {
    return requestToJson(request)
        .flatMap((json) =>
            validateJsonKeys(json, ['name', 'userId'])) // Add required fields
        .flatMap((json) => decodeJson(json, ProjectCreateDto.fromJson))
        .flatMap(mapper.fromCreateTo)
        .flatMap(projects.create)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.put('/<projectId>')
  Future<Response> updateProject(Request request, String projectId) async {
    return requestToJson(request)
        .flatMap((json) => validateJsonKeys(json, []))
        .flatMap((json) => decodeJson(json, ProjectUpdateDto.fromJson))
        .flatMap((dto) => mapper.fromUpdateTo(dto, projectId))
        .flatMap(projects.update)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.delete('/<projectId>')
  Future<Response> deleteProject(Request request, String projectId) async {
    return projects
        .delete(projectId)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok('project deleted'))
        .run();
  }

  // Create router using the generate function defined in 'project_service.g.dart'.
  Router get router => _$ProjectServiceRouter(this);
}
