import 'dart:convert';

import 'package:backend_dart/application/service/dto/project_task_dto.dart';
import 'package:backend_dart/application/service/mapper/project_task_dto_mapper.dart';
import 'package:backend_dart/common/request_helper.dart';
import 'package:backend_dart/common/response_helpers.dart';
import 'package:backend_dart/common/validation.dart';
import 'package:backend_dart/domain/repository/project_task_repository.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_router/shelf_router.dart';

part 'project_task_service.g.dart'; // Generated with 'pub run build_runner build'

class ProjectTaskService {
  final ProjectTaskRepository tasks;
  final ProjectTaskDtoMapper mapper = ProjectTaskDtoMapper();
  ProjectTaskService(this.tasks);

  @Route.get('/')
  Future<Response> listTasks(Request request) async {
    return tasks
        .findAll()
        .flatMap(mapper.listTo)
        .map((tasks) => tasks.map((task) => task.toJson()).toList())
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok(jsonEncode(right)))
        .run();
  }

  @Route.get('/<taskId>')
  Future<Response> fetchTask(Request request, String taskId) async {
    return tasks
        .findById(taskId)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.post('/')
  Future<Response> createTask(Request request) async {
    return requestToJson(request)
        .flatMap((json) =>
            validateJsonKeys(json, ['name', 'projectId'])) // Add required fields
        .flatMap((json) => decodeJson(json, ProjectTaskCreateDto.fromJson))
        .flatMap(mapper.fromCreateTo)
        .flatMap(tasks.create)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.put('/<taskId>')
  Future<Response> updateTask(Request request, String taskId) async {
    return requestToJson(request)
        .flatMap((json) => validateJsonKeys(json, []))
        .flatMap((json) => decodeJson(json, ProjectTaskUpdateDto.fromJson))
        .flatMap((dto) => mapper.fromUpdateTo(dto, taskId))
        .flatMap(tasks.update)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.delete('/<taskId>')
  Future<Response> deleteTask(Request request, String taskId) async {
    return tasks
        .delete(taskId)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok('Task deleted'))
        .run();
  }

  Router get router => _$ProjectTaskServiceRouter(this);
}
