import 'dart:convert';

import 'package:backend_dart/application/service/dto/time_entry_dto.dart';
import 'package:backend_dart/application/service/mapper/time_entry_dto_mapper.dart';
import 'package:backend_dart/common/request_helper.dart';
import 'package:backend_dart/common/response_helpers.dart';
import 'package:backend_dart/common/validation.dart';
import 'package:backend_dart/domain/repository/time_entry_repository.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_router/shelf_router.dart';

part 'time_entry_service.g.dart'; // Generated with 'pub run build_runner build'

class TimeEntryService {
  final TimeEntryRepository entries;
  final TimeEntryDtoMapper mapper = TimeEntryDtoMapper();
  TimeEntryService(this.entries);

  @Route.get('/')
  Future<Response> listEntries(Request request) async {
    return entries
        .findAll()
        .flatMap(mapper.listTo)
        .map((entries) => entries.map((entry) => entry.toJson()).toList())
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok(jsonEncode(right)))
        .run();
  }

  @Route.get('/<entryId>')
  Future<Response> fetchEntry(Request request, String entryId) async {
    return entries
        .findById(entryId)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.post('/')
  Future<Response> createEntry(Request request) async {
    return requestToJson(request)
        .flatMap((json) => validateJsonKeys(
            json, ['startTime', 'endTime', 'userId', 'projectId']))
        .flatMap((json) => decodeJson(json, TimeEntryCreateDto.fromJson))
        .flatMap(mapper.fromCreateTo)
        .flatMap(entries.create)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.put('/<entryId>')
  Future<Response> updateEntry(Request request, String entryId) async {
    return requestToJson(request)
        .flatMap((json) => validateJsonKeys(json, []))
        .flatMap((json) => decodeJson(json, TimeEntryUpdateDto.fromJson))
        .flatMap((dto) => mapper.fromUpdateTo(dto, entryId))
        .flatMap(entries.update)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.delete('/<entryId>')
  Future<Response> deleteEntry(Request request, String entryId) async {
    return entries
        .delete(entryId)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok('Entry deleted'))
        .run();
  }

  Router get router => _$TimeEntryServiceRouter(this);
}
