import 'dart:convert';

import 'package:backend_dart/application/service/dto/user_dto.dart';
import 'package:backend_dart/application/service/mapper/user_dto_mapper.dart';
import 'package:backend_dart/common/request_helper.dart';
import 'package:backend_dart/common/response_helpers.dart';
import 'package:backend_dart/common/validation.dart';
import 'package:backend_dart/domain/repository/user_repository.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_router/shelf_router.dart';

part 'user_service.g.dart'; // generated with 'pub run build_runner build'

class UserService {
  final UserRepository users;
  final UserDtoMapper mapper = UserDtoMapper();
  UserService(this.users);

  @Route.get('/')
  Future<Response> listUsers(Request request) async {
    return users
        .findAll()
        .flatMap(mapper.listTo)
        .map((users) => users.map((user) => user.toJson()).toList())
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok(jsonEncode(right)))
        .run();
  }

  @Route.get('/<userId>')
  Future<Response> fetchUser(Request request, String userId) async {
    return users
        .findById(userId)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.post('/')
  Future<Response> createUser(Request request) async {
    return requestToJson(request)
        .flatMap(
            (json) => validateJsonKeys(json, ['name', 'email', 'password']))
        .flatMap((json) => decodeJson(json, UserCreateDto.fromJson))
        .flatMap(mapper.fromCreateTo)
        .flatMap(users.create)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.put('/<userId>')
  Future<Response> updateUser(Request request, String userId) async {
    return requestToJson(request)
        .flatMap((json) => validateJsonKeys(json, []))
        .flatMap((json) => decodeJson(json, UserUpdateDto.fromJson))
        .flatMap((dto) => mapper.fromUpdateTo(dto, userId))
        .flatMap(users.update)
        .flatMap(mapper.to)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => ResponseHelpers.jsonOk(right.toJson()))
        .run();
  }

  @Route.delete('/<userId>')
  Future<Response> deleteUser(Request request, String userId) async {
    return users
        .delete(userId)
        .match((left) => ResponseHelpers.fromError(left),
            (right) => Response.ok('user deleted'))
        .run();
  }

  // Create router using the generate function defined in 'userservice.g.dart'.
  Router get router => _$UserServiceRouter(this);
}
