import 'package:backend_dart/common/error_on_null.dart';
import 'package:backend_dart/domain/data/project_data_source.dart';
import 'package:backend_dart/domain/entities/project.dart';
import 'package:backend_dart/domain/errors/app_error.dart';
import 'package:backend_dart/domain/interface/error.dart';
import 'package:backend_dart/infrastructure/persistence/db/client.dart';
import 'package:backend_dart/infrastructure/persistence/db/model.dart';
import 'package:backend_dart/infrastructure/persistence/db/prisma.dart';
import 'package:backend_dart/infrastructure/persistence/mapper/project_dbo_mapper.dart';
import 'package:fpdart/fpdart.dart';
import 'package:orm/orm.dart';
import 'package:uuid/uuid.dart';

class PrismaProjectDataSource implements ProjectDataSource {
  final PrismaClient prisma;
  final ProjectDboMapper mapper = ProjectDboMapper();

  PrismaProjectDataSource(this.prisma);

  @override
  TaskEither<IError, Project> create(ProjectCreate project) {
    return mapper
        .fromCreatetoDbo(project)
        .flatMap((projectDbo) => TaskEither.tryCatch(
              () async {
                return await prisma.projectDbo.create(
                  data: PrismaUnion.$1(projectDbo),
                );
              },
              (error, _) => AppError.databaseError(
                message: 'Failed to create project: ${error.toString()}',
              ),
            ))
        .flatMap(mapper.fromDbo);
  }

  @override
  TaskEither<IError, Project> findById(String id) {
    return TaskEither<IError, ProjectDbo?>.tryCatch(
      () async {
        final project = await prisma.projectDbo
            .findUnique(where: ProjectDboWhereUniqueInput(id: id));
        return project;
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to find project by ID: ${error.toString()}',
      ),
    )
        .flatMap(
            errorOnNull(AppError.notFound('Project with ID $id not found')))
        .flatMap(mapper.fromDbo);
  }

  @override
  TaskEither<IError, List<Project>> findByUserId(String userId) {
    return TaskEither<IError, Iterable<ProjectDbo>>.tryCatch(
      () async {
        final projects = await prisma.projectDbo.findMany(
          where: ProjectDboWhereInput(
              userId: PrismaUnion.$1(StringFilter(
            equals: PrismaUnion.$1(userId),
          ))),
        );
        return projects;
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to find projects by user ID: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, Project> update(ProjectUpdate project) {
    return mapper
        .fromUpdateToDbo(project)
        .flatMap(
          (projectDbo) => TaskEither.tryCatch(
            () async {
              return await prisma.projectDbo.update(
                data: PrismaUnion.$1(projectDbo),
                where: ProjectDboWhereUniqueInput(id: project.id),
              );
            },
            (error, _) => AppError.databaseError(
              message: 'Failed to update project: ${error.toString()}',
            ),
          ),
        )
        .flatMap(
          errorOnNull(
            AppError.notFound(
              'Project with ID ${project.id} not found',
            ),
          ),
        )
        .flatMap(mapper.fromDbo);
  }

  @override
  TaskEither<IError, Project> delete(String id) {
    return TaskEither<IError, ProjectDbo?>.tryCatch(
      () async {
        return await prisma.projectDbo
            .delete(where: ProjectDboWhereUniqueInput(id: id));
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to delete project: ${error.toString()}',
      ),
    )
        .flatMap(
          errorOnNull(
            AppError.notFound(
              'Project with ID $id not found',
            ),
          ),
        )
        .flatMap(mapper.fromDbo);
  }

  @override
  TaskEither<IError, List<Project>> findAll() {
    return TaskEither<IError, Iterable<ProjectDbo>>.tryCatch(
      () async => await prisma.projectDbo.findMany(),
      (error, _) => AppError.databaseError(
        message: 'Failed to fetch all projects: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, String> generateId() {
    return TaskEither.tryCatch(
      () async {
        var uuid = Uuid();
        do {
          final id = uuid.v4();
          final project = await prisma.projectDbo.findUnique(
            where: ProjectDboWhereUniqueInput(id: id),
          );
          if (project == null) return id;
        } while (true);
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to generate ID: ${error.toString()}',
      ),
    );
  }
}
