import 'package:backend_dart/common/error_on_null.dart';
import 'package:backend_dart/domain/data/project_task_data_source.dart';
import 'package:backend_dart/domain/entities/project_task.dart';
import 'package:backend_dart/domain/errors/app_error.dart';
import 'package:backend_dart/domain/interface/error.dart';
import 'package:backend_dart/infrastructure/persistence/db/client.dart';
import 'package:backend_dart/infrastructure/persistence/db/model.dart';
import 'package:backend_dart/infrastructure/persistence/db/prisma.dart';
import 'package:backend_dart/infrastructure/persistence/mapper/project_task_dbo_mapper.dart';
import 'package:fpdart/fpdart.dart';
import 'package:orm/orm.dart';
import 'package:uuid/uuid.dart';

class PrismaProjectTaskDataSource implements ProjectTaskDataSource {
  final PrismaClient prisma;
  final ProjectTaskDboMapper mapper = ProjectTaskDboMapper();

  PrismaProjectTaskDataSource(this.prisma);

  @override
  TaskEither<IError, ProjectTask> create(ProjectTaskCreate task) =>
      mapper.fromCreateTo(task).flatMap(
            (taskDbo) => TaskEither<IError, ProjectTaskDbo>.tryCatch(
              () async {
                final createdTask = await prisma.projectTaskDbo.create(
                  data: PrismaUnion.$1(taskDbo),
                );
                return createdTask;
              },
              (error, _) => AppError.databaseError(
                message: 'Failed to create project task: ${error.toString()}',
              ),
            ).flatMap(mapper.from),
          );

  @override
  TaskEither<IError, ProjectTask> findById(String id) {
    return TaskEither<IError, ProjectTaskDbo?>.tryCatch(
      () async {
        final task = await prisma.projectTaskDbo
            .findUnique(where: ProjectTaskDboWhereUniqueInput(id: id));
        return task;
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to find project task by ID: ${error.toString()}',
      ),
    )
        .flatMap(errorOnNull(AppError.notFound(
          "Project task with id $id not found",
        )))
        .flatMap(mapper.from);
  }

  @override
  TaskEither<IError, List<ProjectTask>> findByProjectId(String projectId) {
    return TaskEither<IError, Iterable<ProjectTaskDbo>>.tryCatch(
      () async => await prisma.projectTaskDbo.findMany(
        where: ProjectTaskDboWhereInput(projectId: PrismaUnion.$2(projectId)),
      ),
      (error, _) => AppError.databaseError(
        message:
            'Failed to fetch tasks for project $projectId: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, ProjectTask> update(ProjectTaskUpdate task) => mapper
      .fromUpdateTo(task)
      .flatMap(
        (taskDbo) => TaskEither.tryCatch(
          () async {
            final updatedTask = await prisma.projectTaskDbo.update(
              data: PrismaUnion.$1(taskDbo),
              where: ProjectTaskDboWhereUniqueInput(id: task.id),
            );
            return updatedTask;
          },
          (error, _) => AppError.databaseError(
            message: 'Failed to update project task: ${error.toString()}',
          ),
        ),
      )
      .flatMap(
        errorOnNull(AppError.notFound("Project task not found")),
      )
      .flatMap(mapper.from);

  @override
  TaskEither<IError, ProjectTask> delete(String id) {
    return TaskEither<IError, ProjectTaskDbo?>.tryCatch(
      () async {
        return await prisma.projectTaskDbo.delete(
          where: ProjectTaskDboWhereUniqueInput(id: id),
        );
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to delete project task: ${error.toString()}',
      ),
    )
        .flatMap(
          errorOnNull(
            AppError.notFound(
              'Project task with ID $id not found',
            ),
          ),
        )
        .flatMap(mapper.from);
  }

  @override
  TaskEither<IError, List<ProjectTask>> findAll() {
    return TaskEither<IError, Iterable<ProjectTaskDbo>>.tryCatch(
      () async => await prisma.projectTaskDbo.findMany(),
      (error, _) => AppError.databaseError(
        message: 'Failed to fetch all project tasks: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, String> generateId() {
    return TaskEither.tryCatch(
      () async {
        var uuid = Uuid();
        do {
          final id = uuid.v4();
          final task = await prisma.projectTaskDbo.findUnique(
            where: ProjectTaskDboWhereUniqueInput(id: id),
          );
          if (task == null) return id;
        } while (true);
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to generate ID: ${error.toString()}',
      ),
    );
  }
}
