import 'package:backend_dart/common/error_on_null.dart';
import 'package:backend_dart/domain/data/time_entry_data_source.dart';
import 'package:backend_dart/domain/entities/time_entry.dart';
import 'package:backend_dart/domain/errors/app_error.dart';
import 'package:backend_dart/domain/interface/error.dart';
import 'package:backend_dart/infrastructure/persistence/db/client.dart';
import 'package:backend_dart/infrastructure/persistence/db/model.dart';
import 'package:backend_dart/infrastructure/persistence/db/prisma.dart';
import 'package:backend_dart/infrastructure/persistence/mapper/time_entry_dbo_mapper.dart';
import 'package:fpdart/fpdart.dart';
import 'package:orm/orm.dart';
import 'package:uuid/uuid.dart';

class PrismaTimeEntryDataSource implements TimeEntryDataSource {
  final PrismaClient prisma;
  final TimeEntryDboMapper mapper = TimeEntryDboMapper();

  PrismaTimeEntryDataSource(this.prisma);

  @override
  TaskEither<IError, TimeEntry> create(TimeEntryCreate timeEntry) =>
      mapper.fromCreateTo(timeEntry).flatMap(
            (timeEntryDbo) => TaskEither<IError, TimeEntryDbo>.tryCatch(
              () async {
                final createdEntry = await prisma.timeEntryDbo.create(
                  data: PrismaUnion.$1(timeEntryDbo),
                );
                return createdEntry;
              },
              (error, _) => AppError.databaseError(
                message: 'Failed to create time entry: ${error.toString()}',
              ),
            ).flatMap(mapper.from),
          );

  @override
  TaskEither<IError, TimeEntry> findById(String id) {
    return TaskEither<IError, TimeEntryDbo?>.tryCatch(
      () async {
        final timeEntry = await prisma.timeEntryDbo
            .findUnique(where: TimeEntryDboWhereUniqueInput(id: id));
        return timeEntry;
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to find time entry by ID: ${error.toString()}',
      ),
    )
        .flatMap(errorOnNull(AppError.notFound(
          "Time entry with id $id not found",
        )))
        .flatMap(mapper.from);
  }

  @override
  TaskEither<IError, List<TimeEntry>> findByUserId(String userId) {
    return TaskEither<IError, Iterable<TimeEntryDbo>>.tryCatch(
      () async => await prisma.timeEntryDbo.findMany(
        where: TimeEntryDboWhereInput(userId: PrismaUnion.$2(userId)),
      ),
      (error, _) => AppError.databaseError(
        message:
            'Failed to fetch time entries for user $userId: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, List<TimeEntry>> findByProjectId(String projectId) {
    return TaskEither<IError, Iterable<TimeEntryDbo>>.tryCatch(
      () async => await prisma.timeEntryDbo.findMany(
        where: TimeEntryDboWhereInput(projectId: PrismaUnion.$2(projectId)),
      ),
      (error, _) => AppError.databaseError(
        message:
            'Failed to fetch time entries for project $projectId: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, TimeEntry> update(TimeEntryUpdate timeEntry) => mapper
      .fromUpdateTo(timeEntry)
      .flatMap(
        (timeEntryDbo) => TaskEither.tryCatch(
          () async {
            final updatedEntry = await prisma.timeEntryDbo.update(
              data: PrismaUnion.$1(timeEntryDbo),
              where: TimeEntryDboWhereUniqueInput(id: timeEntry.id),
            );
            return updatedEntry;
          },
          (error, _) => AppError.databaseError(
            message: 'Failed to update time entry: ${error.toString()}',
          ),
        ),
      )
      .flatMap(
        errorOnNull(AppError.notFound("Time entry not found")),
      )
      .flatMap(mapper.from);

  @override
  TaskEither<IError, TimeEntry> delete(String id) {
    return TaskEither<IError, TimeEntryDbo?>.tryCatch(
      () async {
        return await prisma.timeEntryDbo.delete(
          where: TimeEntryDboWhereUniqueInput(id: id),
        );
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to delete time entry: ${error.toString()}',
      ),
    )
        .flatMap(
          errorOnNull(AppError.notFound(
            "Time entry with id $id not found",
          )),
        )
        .flatMap(mapper.from);
  }

  @override
  TaskEither<IError, List<TimeEntry>> findAll() {
    return TaskEither<IError, Iterable<TimeEntryDbo>>.tryCatch(
      () async => await prisma.timeEntryDbo.findMany(),
      (error, _) => AppError.databaseError(
        message: 'Failed to fetch all time entries: ${error.toString()}',
      ),
    ).flatMap(mapper.listFrom);
  }

  @override
  TaskEither<IError, String> generateId() {
    return TaskEither.tryCatch(
      () async {
        var uuid = Uuid();
        do {
          final id = uuid.v4();
          final entry = await prisma.timeEntryDbo.findUnique(
            where: TimeEntryDboWhereUniqueInput(id: id),
          );
          if (entry == null) return id;
        } while (true);
      },
      (error, _) => AppError.databaseError(
        message: 'Failed to generate ID: ${error.toString()}',
      ),
    );
  }
}
